/* 
 * noIdle v1.1b (date: 28-Feb-2002)
 *
 * Program that don't let user process get more than a specific idle time.
 * 
 * (C) 2002 - Jos Martins (jemart@student.dei.uc.pt)
 * 
 * */

#include "define.h"

static int pipe_fd[2];
static FILE *log_file;

char log_file_name[MAX_LINE]=LOG_FILE;

int options=(0|OPC_DEFAULT);

int cicle_time=CICLE_TIME, idle_time=IDLE_TIME;

int main(int argc, char *argv[]){
  
  int len, offs=0, pids_offs=0, flag=0, child_pid, pid_file;
  char processes[MAX_LINE]="", pids[MAX_LINE]="", user[MAX_LINE]="", token[MAX_LINE]="",pid_token[MAX_LINE]="",log_msg[MAX_LINE],star_line[MAX_LINE]="",pid_char[MAX_LINE];
  struct timeval timeout;
  process c_process,pid_process;
  
  if(processArgs(argc,argv)==EXIT_FAILURE){
    panicExit(0);
  }
  
  /* Normal daemonization */
  if((child_pid=(int)fork())!=0){
    return EXIT_SUCCESS;
  }
  
  if((errno==EAGAIN) || (errno==ENOMEM)){
    perror("noIdle daemonization error");
    return EXIT_FAILURE;
  }
  
  /* chdir() */
  if(chdir(WORK_DIR)==-1){
    perror("noIdle chdir error");
    return EXIT_FAILURE;
  }
  
  /* create file with pid */
  /* create noIdle.pid */
  (void)sprintf(pid_char,"%s/noIdle.pid",WORK_DIR);
  pid_file=open(pid_char,O_WRONLY|O_CREAT|O_EXCL,S_IRUSR|S_IWUSR);
  if(pid_file==-1){
    switch(errno){
      case EEXIST:{printf("lockfile present. noIdle already running.\n(last session broken? try removing noIdle.pid\n"); return EXIT_FAILURE;}break;
      default:{perror("noIdle lockfile error"); return EXIT_FAILURE;}break;
    }
  }
  (void)sprintf(pid_char,"%d\n",getpid());
  if(write(pid_file,pid_char,strlen(pid_char))<=0){
    perror("noIdle error:");
  };
  close(pid_file);
  
  (void)firstAssignments();
  
  if( (options&OPC_HELP)==OPC_HELP ){
    printf("noIdle version %s (build:%s)\n",VERSION,BUILD);
    printf("usage: %s [OPTIONS]\n\nOptions:\n  --log_file=<yes|no>           Create (or not) a logfile\n  --log_filename <filename>     Logfile into <filename>\n  --idle_time n                 specify 'n' idle time in minutes (default=%d)\n  --cicle_time n                specify 'n' time in seconds between cicles\n                                (default=%d)\n  --help                        this message\n  --version                     print version.\n",argv[0],IDLE_TIME,CICLE_TIME);
    panicExit(0);
  };
  
  if( (options&OPC_VERSAO)==OPC_VERSAO ){
    printf("noIdle version %s (build:%s)\n",VERSION,BUILD);
    panicExit(0);
  };

  printMsg(DONOT_DATE,"*************************************************");
  (void)memset(star_line,(int)'\0',MAX_LINE);
  sprintf(star_line,"noIdle started (pid:%d)",getpid());
  printMsg(DO_DATE,star_line);
  
  if(processCmd(CMD_WHOAMI,user,NULL)==EXIT_FAILURE){
    printMsg(DO_DATE,"Error processing whoami comand\n");
  };
  len=(int)strlen(user);
  user[len-1]='\0';
  
  for(;;){
#ifdef DEBUG
    printMsg(DO_DATE,"cicle...");
#endif
    
    /* cicle timeout */
    timeout.tv_sec=cicle_time;
    timeout.tv_usec=0;
  
    (void)memset(processes,(int)'\0',(size_t)MAX_LINE);
    /* reads all processes and pids*/
    if(processCmd(CMD_W,user,processes)==EXIT_FAILURE){
      printMsg(DO_DATE,"Error processing who comand\n");
      panicExit(0);
    };
    
    (void)memset(pids,(int)'\0',(size_t)MAX_LINE);
    if(processCmd(CMD_PS,user,pids)==EXIT_FAILURE){
      printMsg(DO_DATE,"Error processing ps comand\n");
      panicExit(0);
    };
    
    offs=0;
    /* for each process, check if is idle... */
    for(;;){
      /* catch process information in processes */
      (void)memset(&c_process,(int)'\0',(size_t)sizeof(process));
      (void)memset(token,(int)'\0',(size_t)MAX_LINE);
      
      /* time */
      if(getToken(2,token,processes,&offs)==EXIT_FAILURE){
        printMsg(DO_DATE,"Error in get token #2 from processes\n");
	panicExit(0);
      };
      
      c_process.idle=checkIdle(token);
      
      if(c_process.idle==IDLE){
      
        /* tty */
        if(getToken(1,c_process.tty,processes,&offs)==EXIT_FAILURE){
          printMsg(DO_DATE,"Error in get token #1 from processes\n");
          panicExit(0);
        };
      
        /* name */
        if(getToken(3,c_process.name,processes,&offs)==EXIT_FAILURE){
          printMsg(DO_DATE,"Error in get token #3 from processes\n");
	  panicExit(0);
	};
        
        (void)memset(log_msg,(int)'\0',(size_t)MAX_LINE);
	sprintf(log_msg,"cmd=%s tty=%s time=%s is idle",c_process.name,c_process.tty,token);
	printMsg(DO_DATE,log_msg);
	
        /* retira o pid e mata */
	pids_offs=0;
	flag=0;
        for(;;){
          (void)memset(&pid_process,(int)'\0',(size_t)sizeof(process));
          (void)memset(pid_token,(int)'\0',(size_t)MAX_LINE);
          /* get name */
	  (void)getToken(2,pid_process.name,pids,&pids_offs);
          /* get tty */
	  (void)getToken(0,pid_process.tty,pids,&pids_offs);
          if( (strcmp(c_process.tty,pid_process.tty)==0) && (strcmp(c_process.name,pid_process.name)==0) ){
            /* get pid */
            (void)getToken(1,pid_token,pids,&pids_offs);
	    c_process.pid=atoi(pid_token);
            (void)memset(log_msg,(int)'\0',(size_t)MAX_LINE);
	    sprintf(log_msg,"  killing cmd=%s tty=%s pid=%d",c_process.name,c_process.tty,c_process.pid);
	    printMsg(DO_DATE,log_msg);

            /* kill */
	    kill(c_process.pid,SIGINT);
	    kill(c_process.pid,SIGQUIT);
	    kill(c_process.pid,SIGKILL);
	    flag=1;
	    break;
	  };
	  if(flag==1)
            break;
          /* muda de linha */
          while(pids[pids_offs++]!='\n'){};
          if(pids[pids_offs]=='\0')
            break;
	};
	
      };
      
      /* change line */
      while(processes[offs++]!='\n'){};
      if(processes[offs]=='\0')
        break;
    }; /* end of one process. */
    
    (void)select(0,(fd_set*)'\0',(fd_set*)'\0',(fd_set*)'\0',&timeout);
    if(errno==EINTR){
      printMsg(DO_DATE,"Error in select. (big cicle)");
      panicExit(0);
    };
  
  };/* end of CICLE_TIME */
  
  return EXIT_SUCCESS;
};

int processArgs(int arg_num,char *arg_array[]){
  int i=0;
  
  while(++i<arg_num){
    if(strcmp(arg_array[i],"--version")==0){
      options|=OPC_VERSAO;
      continue;
    };

    if(strcmp(arg_array[i],"--help")==0){
      options|=OPC_HELP;
      continue;
    };

    if(strcmp(arg_array[i],"--log_file=yes")==0){
      options|=OPC_LOGFILE;
      continue;
    };
    
    if(strcmp(arg_array[i],"--log_file=no")==0){
      options&=(~OPC_LOGFILE);
      continue;
    };
    
    if(strcmp(arg_array[i],"--log_filename")==0){
      if(++i==arg_num)
	return EXIT_FAILURE;
      strcpy(log_file_name,arg_array[++i]);
      continue;
    };

    if(strcmp(arg_array[i],"--cicle_time")==0){
      if(++i==arg_num)
	return EXIT_FAILURE;
      cicle_time=atoi(arg_array[++i]);
      continue;
    };

    if(strcmp(arg_array[i],"--idle_time")==0){
      if(++i==arg_num)
	return EXIT_FAILURE;
      idle_time=atoi(arg_array[++i]);
      continue;
    };

    printf("Wrong usage.\n  usage: %s [OPTIONS]\n",arg_array[0]);
    return EXIT_FAILURE;
  };
  
  return EXIT_SUCCESS;
};

int firstAssignments(){
  if(assignSignals()==EXIT_FAILURE){
    printf("error assigning signals...\n");
    return EXIT_FAILURE;
  };
  
  /* opens pipes */
  if( pipe(pipe_fd)==-1){
    perror("Error making pipe");
    return EXIT_FAILURE;
  };

  /* open file */
/*  if( (options&OPC_LOGFILE)==OPC_LOGFILE ){
    log_file=fopen(log_file_name,LOG_PERM);
    if(log_file==NULL){
      perror("Error opening log file");
      panicExit(0);
    };
  }else{
    log_file=NULL;
  };*/

  return EXIT_SUCCESS;
};

int processCmd(int cmd, char *arg0, char *arg1){
  int child,status;
  
  child=fork();
  if(child==-1){
    /* error */
    printMsg(DO_DATE,"Error creating child\n");
    return EXIT_FAILURE;
  }else if(child!=0){
    /* father */
    if(arg1==NULL)
      (void)read_pipe_fd(arg0);
    else
      (void)read_pipe_fd(arg1);
    
    if( child!=wait(&status) ){
      printMsg(DO_DATE,"wrong child!!!\n");
      return EXIT_FAILURE;
    };
    return EXIT_SUCCESS;
  };
  /* child */
  
  if( dup2(pipe_fd[1],STDOUT_FILENO)==-1 ){
    /* error */
    printMsg(DO_DATE,"Error creating pipe\n");
    return EXIT_FAILURE;
  };
  
  if( cmd==CMD_WHOAMI )
    (void)execl("/usr/bin/whoami","whoami",NULL);
  if( cmd==CMD_W )
    (void)execl("/usr/bin/w","w","-fsh",arg0,NULL);
  if( cmd==CMD_PS )
    (void)execl("/bin/ps","ps","-U",arg0,"-ho","tty,pid,cmd",NULL);
  printMsg(DO_DATE,"Error in processCmd\n");
  return EXIT_FAILURE;
};

int read_pipe_fd(char *str){
  
  int select_val;
  ssize_t count;
  struct timeval rtimeout;
  fd_set fdset;
  
  (void)memset(str,(int)'\0',MAX_LINE);
    
  /* read timeout of 5sec */
  rtimeout.tv_sec=5;
  rtimeout.tv_usec=0;
  
  /* reset fd_set */
  FD_ZERO(&fdset);
  FD_SET(pipe_fd[0],&fdset);
  /* select */
  select_val=select(pipe_fd[0]+1,&fdset,(fd_set*)'\0',(fd_set*)'\0',&rtimeout);
  if(select_val==1){
    count=read(pipe_fd[0],str,(size_t)MAX_LINE);
  }else if(select_val==-1){
    printMsg(DO_DATE,"Error in select in main\n");
    return EXIT_FAILURE;
  };

  return EXIT_SUCCESS;
};

int checkIdle(char *str){
  int time1=0, time2=0, strLen=0;
#ifdef DEBUG
  char toto[MAX_LINE];
#endif
  char c=' ';
  
  strLen=(int)strlen(str);
  if(sscanf(str,"%d:%d",&time1,&time2)!=2)
    if(sscanf(str,"%d.%d%c",&time1,&time2,&c)!=3)
      return IDLE;
  
  c=str[strLen-1];

#ifdef DEBUG
  (void)sprintf(toto,"got translation t1=%d t2=%d c=%c from %s",time1,time2,c,str);
  printMsg(DO_DATE,toto);
#endif
  
  if(isdigit((int)c)!=0){
  /* min:sec */
    if(time1>=idle_time)
      return IDLE;
    else
      return NOIDLE;
  }else if(c=='s'){
  /* sec:usec */
    return NOIDLE;
  }else if(c=='m'){
  /* h:min */
    return IDLE;
  };
  
  return IDLE;
};

int getToken(int tokNum, char *tok, char *str, int *offs){
  int tokOffs=0, cpOffs=0;
  
  /* get correct token */
  while((tokNum--)!=0){
    while(str[*offs+tokOffs]!=' '){
      tokOffs++;
      if(str[*offs+tokOffs]=='\n')
        return EXIT_FAILURE;
    };
    while(str[*offs+tokOffs]==' '){tokOffs++;};
  };
  
  /* copy token */
  while(str[*offs+tokOffs]!=' '){
    tok[cpOffs++]=str[*offs+tokOffs++];
    if(tok[cpOffs-1]=='\n'){
      tok[cpOffs-1]='\0';
      return EXIT_SUCCESS;
    }
  };
  tok[tokOffs]='\0';
  return EXIT_SUCCESS;
};

void printMsg(int date,char *msg){
  time_t time_mark;
  FILE *out_stream;
  char out_string[2*MAX_LINE],time_string[MAX_LINE];
  int time_len;

  memset(out_string,(int)'\0',2*MAX_LINE);
  memset(time_string,(int)'\0',MAX_LINE);
  
 /* if it's dated...
   * The objective is only to sprintf the out_string. */
  if(date==DO_DATE){
    (void)time(&time_mark);
    strcpy(time_string,ctime(&time_mark));
    sprintf(out_string,"%s",time_string);
    time_len=(int)strlen(time_string);
    strcpy(out_string+time_len-1," -> ");
    sprintf(out_string+time_len+3,"%s",msg);
  }else{
    sprintf(out_string,msg);
  }
    
  /* if "OPC_LOGFILE" then open file for writing...*/
  if( (options&OPC_LOGFILE)==OPC_LOGFILE ){
    log_file=fopen(log_file_name,LOG_PERM);
#ifdef DEBUG
    fprintf(log_file,"log_file opened...\n");
#endif
    if(log_file==NULL){
      perror("Error opening log file");
      panicExit(0);
    };
  };
  
   /* either log_file or stdout */
  if( (options&OPC_LOGFILE)==OPC_LOGFILE )
    out_stream=log_file;
  else
    out_stream=stdout;
     
  fprintf(out_stream,"%s\n",out_string);
  (void)fflush(out_stream);

  /* A-> log_file==null
   * B-> fclose!=0 
   * (A+B) == !( (!A) * (!B) ) */
  if( (options&OPC_LOGFILE)==OPC_LOGFILE ){
#ifdef DEBUG
    fprintf(log_file,"log_file closed...\n");
#endif
    if(!((log_file!=NULL) && (fclose(log_file)==0)))
      perror("Error closing log file");
  }; 
  return ;
};

int assignSignals(void){
  if (signal(SIGSEGV,panicExit)==SIG_ERR ||
      signal(SIGTERM,panicExit)==SIG_ERR ||
      signal(SIGQUIT,panicExit)==SIG_ERR ||
      signal(SIGINT,panicExit)==SIG_ERR)
    return EXIT_FAILURE;
  else
    return EXIT_SUCCESS;
};
 
void panicExit(int signal){
  char pid_char[MAX_LINE];
  if(signal!=0)
    printMsg(DO_DATE,"GOT PANIC!!!");
  
  if(signal==SIGTERM){
    printMsg(DONOT_DATE,"received SIGTERM.");
  }else if(signal==SIGQUIT){
    printMsg(DONOT_DATE,"received SIGQUIT.");
  }else if(signal==SIGINT){
    printMsg(DONOT_DATE,"received SIGINT.");
  }else if(signal==SIGSEGV){
    printMsg(DONOT_DATE,"received SIGSEGV.");
  };

  printMsg(DO_DATE,"exiting...");

  (void)sprintf(pid_char,"%s/noIdle.pid",WORK_DIR);
  if(unlink(pid_char)==-1){
    perror("noIdle unlink error");
  }
  
  exit(EXIT_SUCCESS);
};
